'use strict';

// Homework 4 -------------------------------------------------------------------------------------

// The Obj class

function Obj() {}

Obj.instVarNames = Object.create(null);

Obj.makeNew = function(/* arg1, ... argN */) {
  var newGuy = new this();
  newGuy._instVars = Object.create(null);
  Object.keys(this.instVarNames).forEach(name => newGuy._instVars[name] = null);
  newGuy.init.apply(newGuy, arguments);
  return newGuy;
};

Obj.superclass = null;

Obj.prototype.init = function() {
  // no op
};

Obj.prototype._checkInstVarName = function(x) {
  if (!this.constructor.instVarNames[x]) {
    throw new Error(x + ' is not a valid instance variable of class ' + this.constructor.name);
  } else {
    return true;
  }
};

Obj.prototype._getInstVar = function(x) {
  return this._checkInstVarName(x) && this._instVars[x];
};

Obj.prototype._setInstVar = function(x, v) {
  this._checkInstVarName(x);
  this._instVars[x] = v;
};

// Translation

function trans(ast) {
  return '"use strict";\n\n' + ast.trans(null);
}

AST.prototype.trans = function(C) {
  throw new TODO('implement translation for ' + this.constructor.name);
}

function transStmts(ss, C, whatToDoWithExpStmt, whatToDoIfNoExpStmt) {
  if (ss.length === 0) {
    return whatToDoIfNoExpStmt;
  }

  var code = ss.slice(0, ss.length - 1).map(s => s.trans(C)).join('') + '\n';
  var lastS = ss[ss.length - 1];
  code += lastS instanceof ExpStmt ?
      whatToDoWithExpStmt + ' ' + lastS.e.trans(C) :
      lastS.trans(C) + whatToDoIfNoExpStmt;
  return code;
};

Program.prototype.trans = function(C) {
  return transStmts(this.ss, C, '', 'null');
};

ExpStmt.prototype.trans = function(C) {
  return this.e.trans(C) + ';';
};

Lit.prototype.trans = function(C) {
  return JSON.stringify(this.primValue);
};

BinOp.prototype.trans = function(C) {
  return '(' + this.e1.trans(C) + this.op + this.e2.trans(C) + ')';
};

Var.prototype.trans = function(C) {
  return this.x;
};

ClassDecl.prototype.trans = function(C) {
  return 'function ' + this.C + '() {}\n' +
      this.C + '.prototype = Object.create(' + this.S + '.prototype);\n' +
      this.C + '.prototype.constructor = ' + this.C + ';\n' +
      this.C + '.makeNew = Obj.makeNew;\n' +
      this.C + '.instVarNames = Object.create(' + this.S + '.instVarNames);\n' +
      this.xs.map(x => this.C + '.instVarNames.' + x + ' = true;').join('\n') +
      this.C + '.superclass = ' + this.S + ';\n' +
      this.C + '.prototype.init = function() {};\n';
};

VarDecl.prototype.trans = function(C) {
  return 'var ' + this.x + ' = ' + this.e.trans(C) + ';';
};

VarAssign.prototype.trans = function(C) {
  return this.x + ' = ' + this.e.trans(C) + ';';
};

MethodDecl.prototype.trans = function(C) {
  var body =
      'function(' + this.xs.join(', ') + ') {\n' +
      this.ss.map(s => s.trans(this.C)).join('\n') +
      'return this;\n' +
      '}';
  return this.C + '.prototype[' + JSON.stringify(this.m) + '] = ' + body + ';';
};

Send.prototype.trans = function(C) {
  return this.erecv.trans(C) + '.' + this.m + '(' + this.es.map(e => e.trans(C)).join(', ') + ')';
};

SuperSend.prototype.trans = function(C) {
  return C + '.superclass.prototype.' + this.m + '.call(this' +
      this.es.map(e => ', ' + e.trans(C)).join('') + ')';
};

New.prototype.trans = function(C) {
  return this.C + '.makeNew(' + this.es.map(e => e.trans(C)).join(', ') + ')';
};

Return.prototype.trans = function(C) {
  return 'return ' + this.e.trans(C) + ';';
};

This.prototype.trans = function(C) {
  return 'this';
};

InstVar.prototype.trans = function(C) {
  return 'this._getInstVar("' + this.x + '")';
};

InstVarAssign.prototype.trans = function(C) {
  return 'this._setInstVar("' + this.x + '", ' + this.e.trans(C) + ');';
};

// Homework 5 -------------------------------------------------------------------------------------

// Declare the classes of "primitive" values

function createPrimitiveClass(name, superclass) {
  function C() {}
  C.prototype = Object.create(superclass.prototype);
  C.prototype.constructor
  C.instVarNames = Object.create(null);
  C.superclass = superclass;
  C.makeNew = function() { throw new Error('cannot create a new ' + name); };
  return C;
}

var Num = createPrimitiveClass('Num', Obj);
var Str = createPrimitiveClass('Str', Obj);
var Null = createPrimitiveClass('Null', Obj);
var Bool = createPrimitiveClass('Bool', Obj);
var True = createPrimitiveClass('True', Bool);
var False = createPrimitiveClass('False', Bool);
var Block = createPrimitiveClass('Block', Obj);

// Modify Send and SuperSend to make them work w/ primitive receivers

function classOf(x) {
  if (x === null) {
    return Null;
  } else if (x === true) {
    return True;
  } else if (x === false) {
    return False;
  } else if (typeof x === 'number') {
    return Num;
  } else if (typeof x === 'string') {
    return Str;
  } else if (x instanceof Function) {
    return Block;
  } else if (x instanceof Obj) {
    return x.constructor;
  } else {
    throw new Error("don't know the class of " + JSON.stringify(x));
  }
}

function send(recv, m, ...args) {
  return classOf(recv).prototype[m].call(recv, ...args);
};

function superSend(superclass, recv, m, ...args) {
  return superclass.prototype[m].call(recv, ...args);
};

Send.prototype.trans = function(C) {
  return 'send(' + this.erecv.trans(C) + ', ' + JSON.stringify(this.m) +
      this.es.map(e => ', ' + e.trans(C)).join('') + ')';
};

SuperSend.prototype.trans = function(C) {
  return 'superSend(' + C + '.superclass, this, ' + JSON.stringify(this.m) +
      this.es.map(e => ', ' + e.trans(C)).join('') + ')';
};

// Add methods to Obj so that the operators will still work

Obj.prototype['+'] = function(that) { return this + that; };
Obj.prototype['-'] = function(that) { return this - that; };
Obj.prototype['*'] = function(that) { return this * that; };
Obj.prototype['/'] = function(that) { return this / that; };
Obj.prototype['%'] = function(that) { return this % that; };
Obj.prototype['>'] = function(that) { return this > that; };
Obj.prototype['<'] = function(that) { return this < that; };
Obj.prototype['=='] = function(that) { return this === that; };
Obj.prototype['!='] = function(that) { return this !== that; };

// Translate binary operations to message sends

BinOp.prototype.trans = function(C) {
  return new Send(this.e1, this.op, [this.e2]).trans(C);
};

// Blocks

var _nextActivationId = 0;

MethodDecl.prototype.trans = function(C) {
  var body =
      'function(' + this.xs.join(', ') + ') {\n' +
      'var _activationId = _nextActivationId++;\n' +
      'try {\n' +
      this.ss.map(s => s.trans(this.C)).join('\n') +
      '} catch (_e) {\n' +
      '  if (_e instanceof _R && _e.activationId === _activationId) {\n' +
      '    return _e.value;\n' +
      '  } else {\n' +
      '    throw _e;\n' +
      '  }\n' +
      '}\n' +
      'return this;\n' +
      '}';
  return this.C + '.prototype[' + JSON.stringify(this.m) + '] = ' + body + ';';
};

class _R {
  constructor(activationId, value) {
    this.activationId = activationId;
    this.value = value;
  }
}

Return.prototype.trans = function(C) {
  return 'throw new _R(_activationId, ' + this.e.trans(C) + ');';
};

Block.prototype.call = function(/* arg1, ..., argn */) {
  return this(...arguments);
};

BlockLit.prototype.trans = function(C) {
  return '((' + this.xs.join(', ') + ') => {' +
      transStmts(this.ss, C, 'return', 'return null') + '\n})';
};
